#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <ctype.h>

extern "C" {
#include <inttypes.h>
#include "ivtv_myth.h"
}

using namespace std;

#include "vbidecoder.h"
#include "vbilut.h"
#include "mythcontext.h"

//
// Teletext section
//

#define printable(c) ((((c) & 0x7F) < 0x20 || ((c) & 0x7F) > 0x7E) ? '.' : ((c) & 0x7F))

#define VBI_STARTBOX 0x0b
#define VBI_ENDBOX 0x0a
#define VBI_SEPARATEDMOSIACS 0x1a
#define VBI_HOLDMOSIACS 0x1e
#define VBI_RELEASEMOSIACS 0x1f
#define VBI_RELEASEMOSIACS 0x1f

TeletextPage *m_serialpage = NULL;

TeletextView::TeletextView()
{
    m_decoder = NULL;

    m_pageinput[0] = '1';
    m_pageinput[1] = '0';
    m_pageinput[2] = '0';

    m_curpage = 0x100;
    m_sidebar = 0;
    m_visible = false;
    m_transparent = false;

    m_tt_rows = 25;
    m_tt_cols = 40;
}

TeletextView::~TeletextView()
{
}

void TeletextView::setDecoder(class VbiDecoder *dec)
{
    // XXX - Lock?
    m_decoder = dec;

    connect(dec, SIGNAL(pageUpdated(int)), this, SLOT(pageUpdated(int)));
}

void TeletextView::setSideBar(int div)
{
    // XXX
    m_sidebar = div;
}

void TeletextView::drawPage(void)
{
    bool mosaic;
    bool conceal;
    bool seperation;
    bool flash;
    bool doubleheight;
    bool blink;
    bool hold;
    
    char last_ch = ' ';
    char ch;
    
    int fgcolor = TTCOLOR_WHITE;
    int newfgcolor = TTCOLOR_WHITE;
    int bgcolor = TTCOLOR_BLACK;
    int newbgcolor = TTCOLOR_BLACK;
    
    TeletextPage *page;

    if (!m_decoder)
        return;

    setForegroundColor(fgcolor);
    setBackgroundColor(bgcolor);

    if (!m_decoder->findPage(m_curpage, &page))
    {
        if (!m_transparent)
            for (int i=1; i<40; ++i)
                drawBackground(i, 0);

        setForegroundColor(TTCOLOR_RED);

        char *str = "Teletext page unavailable";
        for (unsigned int i=0; i<strlen(str); ++i)
        {
            drawCharacter(i+7, 0, str[i]);
        }
        goto out;
    }

    for (int y=0; y < m_tt_rows; ++y)
    {
        //do a redraw check here??

        fgcolor = TTCOLOR_WHITE;
        newfgcolor = TTCOLOR_WHITE;
        bgcolor = TTCOLOR_BLACK;

        setForegroundColor(fgcolor);
        setBackgroundColor(bgcolor);

        mosaic = false;
        seperation = false;
        conceal = false;
        flash = false;
        doubleheight = false;
        blink = false;
        hold = false;
        seperation = false;

        for (int x=0; x < m_tt_cols; ++x)
        {
            setForegroundColor(fgcolor);
            setBackgroundColor(bgcolor);

            ch = page->data[y][x] & 0x7F;
            switch(ch) {
                case 0x00 ... 0x07:	// alpha + foreground color 
                    fgcolor = ch & 7;
                    mosaic = false;
                    conceal = false;
                    goto ctrl;
                case 0x08:		// flash 
                    //flash_state = not xw->blink_on;
                    goto ctrl;
                case 0x09:		// steady 
                    flash = false;
                    goto ctrl;
                case 0x0a:		// end box 
                case 0x0b:		// start box 
                    goto ctrl;
                case 0x0c:		// normal height 
                    doubleheight = 0;
                    goto ctrl;
                case 0x0d:		// double height
                    doubleheight = y < m_tt_rows-1;
                    goto ctrl;
                case 0x10 ... 0x17:	// graphics + foreground color 
                    fgcolor = ch & 7;
                    mosaic = true;
                    conceal = false;
                    goto ctrl;
                case 0x18:		// conceal display 
                    conceal = false; //not xw->reveal;
                    goto ctrl;
                case 0x19:		// contiguous graphics
                    seperation = false;
                    goto ctrl;
                case 0x1a:		// separate graphics 
                    seperation = true;
                    goto ctrl;
                case 0x1c:		// black background
                    bgcolor = TTCOLOR_BLACK;
                    goto ctrl;
                case 0x1d:		// new background
                    bgcolor = fgcolor;
                    goto ctrl;
                case 0x1e:		// hold graphics 
                    hold = true;
                    goto ctrl;
                case 0x1f:		// release graphics
                    hold = false;
                    goto ctrl;
                case 0x0e:		// SO (reserved, double width)
                case 0x0f:		// SI (reserved, double size)
                case 0x1b:		// ESC (reserved)
                    ch = ' ';
                    break;
                ctrl:
                    ch = ' ';
                    if (hold && mosaic)
                        ch = last_ch;
                    break;
                case 0x80 ... 0x9f:	// these aren't used
                    ch = ' ' ;//BAD_CHAR;
                        break;

                    default:		// mapped to selected font
                        break;
                }
                
                newfgcolor = fgcolor;
                newbgcolor = bgcolor;

                if (blink)
                {
                    //XXX TODO 
                    //newfgcolor |= 8;                
                }                    

                if (conceal)
                    //newfgcolor = newbgcolor;
                
                setForegroundColor(newfgcolor);
                setBackgroundColor(newbgcolor);

                if (!m_transparent) drawBackground(x, y);

                if ((mosaic) && (ch < 0x40 || ch > 0x5F))
                {
                    setBackgroundColor(newfgcolor);
                    drawMosaic(x, y, ch);
                } 
                else drawCharacter(x, y, ch);
        }
    }
out:
    drawCharacter(1, 0, m_pageinput[0]);
    drawCharacter(2, 0, m_pageinput[1]);
    drawCharacter(3, 0, m_pageinput[2]);
}

void TeletextView::keyPress(enum TTKey key)
{
    bool numeric_input = false;
    int curpage_bcd;
        
    switch (key)
    {
        case TTKey_0 ... TTKey_9:
            numeric_input = true;

            if (m_pageinput[0] == ' ')
                m_pageinput[0] = '0' + key;
            else if (m_pageinput[1] == ' ')
                m_pageinput[1] = '0' + key;
            else if (m_pageinput[2] == ' ')
            {
                m_pageinput[2] = '0' + key;
                // XXX - Signal pagechange.
                m_curpage = ((m_pageinput[0] - '0') * 256) +
                            ((m_pageinput[1] - '0') * 16) +
                            (m_pageinput[2] - '0');
            }
            else
            {
                // XXX - Abort pagechange.
                m_pageinput[0] = '0' + key;
                m_pageinput[1] = ' ';
                m_pageinput[2] = ' ';
            }
            break;
        
        case TTKey_NextPage:
            curpage_bcd++;
            break;
        case TTKey_PrevPage:
            curpage_bcd--;
            break;
        case TTKey_NextSubPage:
            curpage_bcd += 10;
            break;
        case TTKey_PrevSubPage:
            curpage_bcd -= 10;
            break;
        case TTKey_Hold:
            break;
        case TTKey_Transparent:
            m_transparent = !m_transparent;
            break;
    }
 
    // XXX - Wrap?
    if (m_curpage < 0x100)
        m_curpage = 0x100;
    if (m_curpage > 0x899)
        m_curpage = 0x899;

    if (!numeric_input)
    {
        m_pageinput[0] = (m_curpage / 256) + '0';
        m_pageinput[1] = ((m_curpage % 256) / 16) + '0';
        m_pageinput[2] = (m_curpage % 16) + '0';
    }

    pageUpdated(m_curpage);
}

//
//Decoder section
//

#define HAMMING84(d) ( ((d & 0x80) >> 4) | ((d & 0x20) >> 3) | ((d & 0x08) >> 2) | ((d & 0x02) >> 1) )
#define BAD_CHAR	0xb8	// substitute for chars with bad parity
#define PIL(day, mon, hour, min) \
	(((day) << 15) + ((mon) << 11) + ((hour) << 6) + ((min) << 0))
#define CC_SIZE 70

uint32_t decodeHamm24(uint8_t d1, uint8_t d2, uint8_t d3)
{
    uint32_t ret = 0;
    ret |= (d1 & 0x04) << 2;
    ret |= (d1 & 0x70) << 3;
    ret |= (d2 & 0x7F) << 4;
    ret |= (d3 & 0x7F) << 5;
    return ret;
}

int hamm8(uint8_t *p, int *err)
{
    int a = hammtab[p[0]];
    *err += a;
    return a & 15;
}

int hamm16(uint8_t *p, int *err)
{
    int a = hammtab[p[0]];
    int b = hammtab[p[1]];
    *err += a;
    *err += b;
    return (a & 15) | (b & 15) * 16;
}

int hamm24(uint8_t *p, int *err)
{
    int e = hamm24par[0][p[0]] ^ hamm24par[1][p[1]] ^ hamm24par[2][p[2]];
    int x = hamm24val[p[0]] + p[1] % 128 * 16 + p[2] % 128 * 2048;

    *err += hamm24err[e];
    return x ^ hamm24cor[e];
}

int chk_parity(uint8_t *p, int n)
{
    int err;

    for (err = 0; n--; p++)
	if (hamm24par[0][*p] & 32)
	    *p &= 0x7f;
	else
	    *p = BAD_CHAR, err++;
    return err;
}

//ccdecode
char    *ratings[] = {"(NOT RATED)","TV-Y","TV-Y7","TV-G","TV-PG","TV-14","TV-MA","(NOT RATED)"};
int     rowdata[] = {11,-1,1,2,3,4,12,13,14,15,5,6,7,8,9,10};
char    *modes[]={"current","future","channel","miscellanious","public service","reserved","invalid","invalid","invalid","invalid"};
char	*specialchar[] = {"®","°","½","¿","(TM)","¢","£","o/~ ","à"," ","è","â","ê","î","ô","û"};
int     lastcode;
int     ccmode=1;               //cc1 or cc2
char    ccbuf[3][256];          //cc is 32 columns per row, this allows for extra characters
int     keywords=0;
char    *keyword[32];

VbiDecoder::VbiDecoder()
{
    m_seendata = false;
    m_waitpage = -1;

    for (int i=0; i<8; ++i) {
        QMutexLocker lock(&m_magazines[i].lock);

        m_magazines[i].pages.clear();
        m_magazines[i].current_page = NULL;
    }
}

VbiDecoder::~VbiDecoder()
{
}

void VbiDecoder::dumpPil(int pil)
{
    int day, mon, hour, min;

    day = pil >> 15;
    mon = (pil >> 11) & 0xF;
    hour = (pil >> 6) & 0x1F;
    min = pil & 0x3F;

    if (pil == PIL(0, 15, 31, 63))
    VERBOSE(VB_VBI,"PDC: Timer-control (no PDC)");
        else if (pil == PIL(0, 15, 30, 63))
            VERBOSE(VB_VBI,"PDC: Recording inhibit/terminate");
        else if (pil == PIL(0, 15, 29, 63))
            VERBOSE(VB_VBI,"PDC: Interruption");
        else if (pil == PIL(0, 15, 28, 63))
            VERBOSE(VB_VBI,"PDC: Continue");
        else if (pil == PIL(31, 15, 31, 63))
            VERBOSE(VB_VBI,"PDC: No time");
        else
            VERBOSE(VB_VBI,QString("PDC: %1, 200X-%2-%3 %4:%5")
            .arg(pil).arg(mon).arg(day).arg(hour).arg(min));
}

int VbiDecoder::decodeVps(uint8_t *buf)
{
    static char pr_label[20];
    static char label[20];
    static int l = 0;
    int cni, pcs, pty, pil;
    int c;
    //unsigned char *buf = p;

    VERBOSE(VB_VBI,"\nVPS:");
    c = vbi_bit_reverse[buf[1]];

    if ((int8_t) c < 0) {
        label[l] = 0;
        memcpy(pr_label, label, sizeof(pr_label));
        l = 0;
    }
    c &= 0x7F;
    label[l] = printable(c);
    l = (l + 1) % 16;
    VERBOSE(VB_VBI, QString(" 3-10: %1 %2 %3 %4 %5 %6 %7 %8 (\"%9\")")
        .arg(buf[0]).arg(buf[1]).arg(buf[2]).arg(buf[3]).arg(buf[4])
        .arg(buf[5]).arg(buf[6]).arg(buf[7]).arg(pr_label));
    pcs = buf[2] >> 6;
    cni = + ((buf[10] & 3) << 10)
          + ((buf[11] & 0xC0) << 2)
          + ((buf[8] & 0xC0) << 0)
          + (buf[11] & 0x3F);
    pil = ((buf[8] & 0x3F) << 14) + (buf[9] << 6) + (buf[10] >> 2);
    pty = buf[12];
    VERBOSE(VB_VBI, QString("CNI: %1 PCS: %2 PTY: %3 ").arg(cni).arg(pcs).arg(pty));
    dumpPil(pil);

    return c & 0xf0;
}

int VbiDecoder::decodeWss(uint8_t *buf)
{
    return 1;
    static const int wss_bits[8] = {
        0, 0, 0, 1, 0, 1, 1, 1 };
    unsigned char parity;
    int wss = 0;
    int i;

    for (i = 0; i < 16; i++) {
        int b1 = wss_bits[buf[i] & 7];
        int b2 = wss_bits[(buf[i] >> 3) & 7];

        if (b1 == b2) return -1;
            wss |= b2 << i;
    }
    parity = wss & 15;
    parity ^= parity >> 2;
    parity ^= parity >> 1;

    fprintf(stderr,"WSS: %s; %s mode; %s color coding;\n"
        "      %s helper; reserved b7=%d; %s\n"
        "      open subtitles: %s; %scopyright %s; copying %s\n",
        formats[wss & 7],
        (wss & 0x10) ? "film" : "camera",
        (wss & 0x20) ? "MA/CP" : "standard",
        (wss & 0x40) ? "modulated" : "no",
        !!(wss & 0x80),
        (wss & 0x0100) ? "have TTX subtitles; " : "",
        subtitles[(wss >> 9) & 3],
        (wss & 0x0800) ? "surround sound; " : "",
        (wss & 0x1000) ? "asserted" : "unknown",
        (wss & 0x2000) ? "restricted" : "not restricted");

    if (!(parity & 1)) return -1;

    return wss;
}

int VbiDecoder::oddParity(uint8_t c)
{
    c ^= (c >> 4);
    c ^= (c >> 2);
    c ^= (c >> 1);

    return c & 1;
}

void VbiDecoder::decodeXds(uint8_t *buf)
{
    char c;

    VERBOSE(VB_VBI,QString("XDS: %1 %2").arg(buf[0]).arg(buf[1]));
    c = oddParity(buf[0]) ? buf[0] & 0x7F : '?';
    c = printable(c);
    putchar(c);
    c = oddParity(buf[1]) ? buf[1] & 0x7F : '?';
    c = printable(c);
    putchar(c);
    putchar('\n');
}

int VbiDecoder::ccDecoder(int b1, int b2, int len)
{
    int x,y, row;

    if ((b1&0x10) && (b2>0x1F)) {
        //codes are always transmitted twice 
        //(apparently not, ignore the second occurance)
        ccmode=((b1>>3)&1)+1;
        len = strlen(ccbuf[ccmode]);

        if (b2 & 0x40) { //preamble address code (row & indent)
            row=rowdata[((b1<<1)&14)|((b2>>5)&1)];
            if (len!=0)
                ccbuf[ccmode][len++]='\n';
                if (b2&0x10) //row contains indent flag
                    for (x=0;x<(b2&0x0F)<<1;x++)
                        ccbuf[ccmode][len++]=' ';
        }
        else {
            switch (b1 & 0x07) {
                case 0x00:      //attribute
                    VERBOSE(VB_VBI, QString("<ATTRIBUTE %1 %2").arg(b1).arg(b2));
                    fflush(stdout);
                    break;
                case 0x01:      //midrow or char
                    switch (b2&0x70) {
                        case 0x20: //midrow attribute change
                            switch (b2&0x0e) {
                                case 0x00: //italics off
                                    strcat(ccbuf[ccmode],"\33[0m ");
                                    break;
                                case 0x0e: //italics on
                                    strcat(ccbuf[ccmode],"\33[36m ");
                                    break;
                            }
                            if (b2&0x01) //underline
                                strcat(ccbuf[ccmode],"\33[4m");
                            else
                                strcat(ccbuf[ccmode],"\33[24m");
                            break;
                        case 0x30: //special character..
                            strcat(ccbuf[ccmode],specialchar[b2&0x0f]);
                            break;
                    }
                    break;
                case 0x04:      //misc
                case 0x05:      //misc + F
                    //printf("ccmode %d cmd %02x\n",ccmode,b2);
                    switch (b2) {
                        case 0x21: //backspace
                            ccbuf[ccmode][len--]=0;
                            break;
                            /* these codes are insignifigant if we're ignoring positioning */
                        case 0x25: //2 row caption
                        case 0x26: //3 row caption
                        case 0x27: //4 row caption
                        case 0x29: //resume direct caption
                        case 0x2B: //resume text display
                        case 0x2C: //erase displayed memory
                            break;
                        case 0x2D: //carriage return
                            if (ccmode==2)
                                break;
                        case 0x2F: //end caption + swap memory
                        case 0x20: //resume caption (new caption)
                            if (!strlen(ccbuf[ccmode]))
                                break;
                            for (x=0;x<strlen(ccbuf[ccmode]);x++)
                                for (y=0;y<keywords;y++)
                                    if (!strncasecmp(keyword[y], ccbuf[ccmode]+x, strlen(keyword[y])))
                                        //printf("%s\n",ccbuf[ccmode]);
                                        /* FALL */
                                        case 0x2A: //text restart
                                        case 0x2E: //erase non-displayed memory
                                            memset(ccbuf[ccmode],0,255);
                                            break;
                    }
                    break;
                case 0x07:      //misc (TAB)
                    for(x=0;x<(b2&0x03);x++)
                        ccbuf[ccmode][len++]=' ';
                        break;
            }
        }
    }
    return 0;
}

int VbiDecoder::decodeCaption(uint8_t *buf, int line)
{
    static int xds_transport = 0;
    char c = buf[0] & 0x7F;
    static char cc[CC_SIZE + 1];
    static int cc_idx;
    VERBOSE(VB_VBI,"decodeCaption");
    
    if (line >=270) {/* in field 2 */
        if (oddParity(buf[0]) && (c >= 0x01 && c <= 0x0F)) {
            decodeXds(buf);
            xds_transport = (c != 0x0F);
        } else if (xds_transport) {
            decodeXds(buf);
        }
    }

    c = oddParity(buf[0]) ? buf[0] & 0x7F : '?';
    c = printable(c);
    if (cc_idx >= CC_SIZE) {
        cc_idx = CC_SIZE - 2;
        memcpy(cc, cc + 2, cc_idx);
    } 
    cc[cc_idx++] = c;
    c = oddParity(buf[1]) ? buf[1] & 0x7F : '?';
    c = printable(c);
    cc[cc_idx++] = c;
    cc[cc_idx] = 0;

    ccDecoder(buf[0], buf[1], cc_idx);

    if (cc_idx == CC_SIZE) {
        VERBOSE(VB_VBI,QString("CC: %1 %2 %3").arg(buf[0]).arg(buf[1]).arg(cc));
        memset(cc, 0, CC_SIZE);
        cc_idx = 0;

    } 

    return 0;
}


void VbiDecoder::reset(void)
{
    // XXX - lock?
    for (int i=0; i<8; ++i)
    {
        QMutexLocker lock(&m_magazines[i].lock);
        m_magazines[i].pages.clear();
        m_magazines[i].current_page = NULL;
        VERBOSE(VB_VBI, QString("resetting magazine %1, %2 pages left")
                .arg(i).arg(m_magazines[i].pages.size()));
    }
}
static uint8_t lang_char[256];

void VbiDecoder::charConversion(uint8_t *buf, int n, int lang)
{
    int c, gfx = 0;

    while (n--)
    {
        if (lang_char[c = *buf])
        {
            if (not gfx || (c & 0xa0) != 0x20)
                *buf = lang_chars[lang + 1][lang_char[c]];
        }
        else if ((c & 0xe8) == 0)
            gfx = c & 0x10;
        buf++;
    }
}

bool VbiDecoder::findPage(int page, TeletextPage **ttpage)
{
    int mag = page / 256;
    if (mag > 8 || mag < 1)
        return false;
    if (mag == 8)
        mag = 0;

    QMutexLocker lock(&m_magazines[mag].lock);

    std::map<int, TeletextPage>::iterator iter;
    iter = m_magazines[mag].pages.find(page % 256);

    if (iter == m_magazines[mag].pages.end()) 
        return false;

    *ttpage = &iter->second;
    return true;
}

#define PG_ACTIVE	0x100	// currently fetching this page

int VbiDecoder::decodeTeletext(uint8_t *buf, int line)
{
    m_seendata = true;
    int err = 0;
    int latin1 = -1;
    uint8_t header = hamm16(buf, &err);
    if (err & 0xf000)
       return -4;
    
    uint8_t magazine = header & 7;
    uint8_t packet = (header >> 3) & 0x1f;

    QMutexLocker lock(&m_magazines[magazine].lock);

    buf += 2;

    switch (packet) {
    
        case 0:
        { // Page Header
            TeletextPage *old_page = m_magazines[magazine].current_page;
            TeletextPage *new_page = NULL;

            int b1, b2, b3, b4;

            b1 = hamm16(buf, &err);	// page number

            b2 = hamm16(buf+2, &err);	// subpage number + flags
            b3 = hamm16(buf+4, &err);	// subpage number + flags
            b4 = hamm16(buf+6, &err);	// language code + more flags
          
            m_serialmode = (b2 + b3 * 256) & 0x3f7f;
            if (m_serialmode)
                for (int i=0; i<8; ++i)
                    if (m_magazines[i].current_page)
                        old_page = m_magazines[i].current_page;


            if (old_page)
            {
                m_magazines[magazine].lock.unlock();
                pageUpdated(old_page->pagenum);
                m_magazines[magazine].lock.lock();
            }

            new_page = &m_magazines[magazine].pages[b1];
            new_page->pagenum =(magazine?: 8) * 256 + b1;
            new_page->lang = "\0\4\2\6\1\5\3\7"[b4 >> 5] + (latin1 ? 0 : 8);
            new_page->flags = b4 & 0x1f;
            new_page->flags |= b3 & 0xc0;
            new_page->flags |= (b2 & 0x80) >> 2;
            new_page->lines = 1;
            new_page->flof = 0;

            //XXX TODO move to text drawing section?
            //charConversion(buf + 8, 32, new_page->lang);

            //if (b1 == 0xff)
                //return 0;
            //if (new_page->pagenum == 0x888)
                //printf("subtitle\n");
                  
            new_page->flags |= PG_ACTIVE;
            memcpy(new_page->data[0]+0, buf, 40);
            memset(new_page->data[0]+40, ' ', sizeof(new_page->data)-40);

            m_magazines[magazine].current_page = new_page;
        }
        break;

    
        case 1 ... 24:
        { // Page Data
            if (!m_magazines[magazine].current_page) {
                return -1;
            }

            TeletextPage *page = m_magazines[magazine].current_page;
            //page->lines |= 1 << packet;
            
            //XXX TODO move to text drawing section
            //charConversoin(buf, 40, page->lang);
            memcpy(page->data[packet], buf, 40);
            pageUpdated(page->pagenum);
        }        
        break;

        case 26:
        {   /* XXX TODO: Level 1.5, 2.5, 3.5
             *      Character location & override.
             * Level 2.5, 3.5
             *      Modifying display attributes.
             * All levels
             *      VCR Programming
             * See 12.3
             */
        }
        break;

        case 27:
        {   // XXX TODO: FLOF data (FastText)
        }
        break;

        case 31:
        {   // XXX TODO private streams
        }
        break;
        
        default:
        {
            // XXX TODO other packet codes...
            //fprintf(stderr,"VbiDecoder: packet ID = %x", packet);
        }
   }
   return 0;
}

void VbiDecoder::processVbiData(uint8_t *buf)
{
        long linemask[2] = { 0, 0 };
        int i, wss, l, id2;
        int field_lines=625; //what is this?
        uint8_t data[43];

        // XXX TODO: nomally we should check on 'itv0' or 'ITV0' 
        // but as we already check the stream type with the first
        // character we will do the check on the remaining chars here...

        if (!memcmp(buf, "tv0", 3)) {
            memcpy(linemask, buf + 3, 8);
            buf += 11;
        }
        else if (!memcmp(buf, "TV0", 3)) {
            linemask[0] = 0xffffffff;
            linemask[1] = 0xf;
            buf += 3;
        }
        else {
            VERBOSE(VB_VBI,"VbiDecoder: unknown VBI data stream");
            return;
        }

        for (i = 0; i < 36; i++) {
            int err = 0;
  
            if (i < 32 && !(linemask[0] & (1 << i))) continue;
            if (i >= 32 && !(linemask[1] & (1 << (i - 32)))) continue;
            l = (i < 18) ? i + 6 : i - 18 + 6 + field_lines;
            id2 = *buf & 0xf;
            switch (id2) {
                case VBI_TYPE_TELETEXT:
                    memcpy(data, buf+1, 42);
                    err = decodeTeletext(data,l);
                    break;
                case VBI_TYPE_CC:
                    err = decodeCaption(buf+1,l);
                    break;
                case VBI_TYPE_VPS:
                    err = decodeVps(buf+1);
                    break;
                case VBI_TYPE_WSS:
                    err = decodeWss(buf+1);
                    break;
            }
            buf += 43;
        }
        return;
}
